# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from typing import Any, Dict


@pytest.fixture
def swh_loader_config() -> Dict[str, Any]:
    return {
        'storage': {
            'cls': 'pipeline',
            'steps': [
                {
                    'cls': 'memory',
                },
            ],
        },
        'deposit': {
            'url': 'https://deposit.softwareheritage.org/1/private',
            'auth': {
                'username': 'user',
                'password': 'pass',
            }
        },
    }



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import requests

from typing import Dict, Optional, Any, Mapping

from swh.model import hashutil

from swh.model.model import (
    Sha1Git, Revision, RevisionType
)

from swh.loader.package.utils import EMPTY_AUTHOR

from swh.loader.package.loader import PackageLoader


def retrieve_sources(url: str) -> Dict[str, Any]:
    response = requests.get(url,
                            allow_redirects=True)
    if response.status_code != 200:
        raise ValueError("Got %d HTTP code on %s",
                         response.status_code, url)

    return json.loads(response.content.decode('utf-8'))


class FunctionalLoader(PackageLoader):
    """Load sources from a sources.json file. This loader is used to load
    sources used by functional package manager (eg. Nix and Guix).

    """
    visit_type = 'functional'

    def __init__(self, url):
        super().__init__(url=url)
        s = retrieve_sources(url)
        self.sources = s['sources']
        self.provider_url = url
        # The revision used to create the sources.json file. For Nix,
        # this revision belongs to the github.com/nixos/nixpkgs
        # repository
        self.revision = s['revision']

    # Note: this could be renamed get_artifacts in the PackageLoader
    # base class.
    def get_versions(self):
        # TODO: try all mirrors and not only the first one. A source
        # can be fetched from several urls, called mirrors. We
        # currently only use the first one, but if the first one
        # fails, we should try the second one and so on.
        return [s['url'][0] for s in self.sources]

    # Note: this could be renamed get_artifact_info in the PackageLoader
    # base class.
    def get_package_info(self, source):
        # TODO: we need to provide the sha256 of the source also
        yield source, {'url': source, 'raw': {'url': source}}

    def resolve_revision_from(
            self, known_artifacts: Dict, artifact_metadata: Dict) \
            -> Optional[bytes]:
        for rev_id, known_artifact in known_artifacts.items():
            known_url = known_artifact['extrinsic']['raw']['url']
            if artifact_metadata['url'] == known_url:
                return rev_id
        return None

    def extra_branches(self) -> Dict[bytes, Mapping[str, Any]]:
        """We add a branch to the snapshot called 'evaluation' pointing to the
        revision used to generate the sources.json file. This revision
        is specified in the sources.json file itself. For the nixpkgs
        origin, this revision is coming from the
        github.com/nixos/nixpkgs repository.

        Note this repository is not loaded explicitly. So, this
        pointer can target a nonexistent revision for a time. However,
        the github and gnu loaders are supposed to load this revision
        and should create the revision pointed by this branch.

        This branch can be used to identify the snapshot associated to
        a Nix/Guix evaluation.

        """
        return {
            b'evaluation': {
                'target_type': 'revision',
                'target': hashutil.hash_to_bytes(self.revision)
            }
        }

    def build_revision(self, a_metadata: Dict, uncompressed_path: str,
                       directory: Sha1Git) -> Optional[Revision]:
        return Revision(
            type=RevisionType.TAR,
            message=b'',
            author=EMPTY_AUTHOR,
            date=None,
            committer=EMPTY_AUTHOR,
            committer_date=None,
            parents=[],
            directory=directory,
            synthetic=True,
            metadata={
                'extrinsic': {
                    'provider': self.provider_url,
                    'when': self.visit_date.isoformat(),
                    'raw': a_metadata,
                },
            }
        )



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.loader.package.loader import PackageLoader


class FakeStorage:
    def origin_add_one(self, origin):
        raise ValueError('We refuse to add an origin')


class FakeStorage2:
    def origin_add_one(self, origin):
        return origin

    def origin_visit_add(self, origin, date, type):
        raise ValueError('We refuse to add an origin visit')


def test_loader_origin_visit_failure(swh_config):
    """Failure to add origin or origin visit should failed immediately

    """
    loader = PackageLoader('some-url')
    loader.storage = FakeStorage()

    actual_load_status = loader.load()
    assert actual_load_status == {'status': 'failed'}

    loader.storage = FakeStorage2()

    actual_load_status2 = loader.load()
    assert actual_load_status2 == {'status': 'failed'}



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


def test_pypi_loader(mocker, swh_app, celery_session_worker, swh_config):
    mock_loader = mocker.patch(
        'swh.loader.package.pypi.loader.PyPILoader.load')
    mock_loader.return_value = {'status': 'eventful'}

    res = swh_app.send_task(
        'swh.loader.package.pypi.tasks.LoadPyPI',
        kwargs=dict(url='some-url'))
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {'status': 'eventful'}



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import hashlib
import logging
import pytest

from swh.model.model import Origin

from swh.loader.core.loader import BaseLoader, DVCSLoader


class DummyLoader:
    def cleanup(self):
        pass

    def prepare(self, *args, **kwargs):
        pass

    def fetch_data(self):
        pass

    def store_data(self):
        pass

    def prepare_origin_visit(self, *args, **kwargs):
        origin = Origin(url='some-url')
        self.origin = origin
        self.origin_url = origin.url
        self.visit_date = datetime.datetime.utcnow()
        self.visit_type = 'git'
        origin_url = self.storage.origin_add_one(origin)
        self.visit = self.storage.origin_visit_add(
            origin_url, self.visit_date, self.visit_type)


class DummyDVCSLoader(DummyLoader, DVCSLoader):
    """Unbuffered loader will send directly to storage new data

    """
    def parse_config_file(self, *args, **kwargs):
        return {
            'max_content_size': 100 * 1024 * 1024,
            'storage': {
                'cls': 'pipeline',
                'steps': [
                    {
                        'cls': 'retry',
                    },
                    {
                        'cls': 'filter',
                    },
                    {
                        'cls': 'memory',
                    },
                ]
            },
        }


class DummyBaseLoader(DummyLoader, BaseLoader):
    """Buffered loader will send new data when threshold is reached

    """
    def parse_config_file(self, *args, **kwargs):
        return {
            'max_content_size': 100 * 1024 * 1024,
            'storage': {
                'cls': 'pipeline',
                'steps': [
                    {
                        'cls': 'retry',
                    },
                    {
                        'cls': 'filter',
                    },
                    {
                        'cls': 'buffer',
                        'min_batch_size': {
                            'content': 2,
                            'content_bytes': 8,
                            'directory': 2,
                            'revision': 2,
                            'release': 2,
                        },
                    },
                    {
                        'cls': 'memory',
                    },
                ]
            },
        }


def test_base_loader():
    loader = DummyBaseLoader()
    result = loader.load()

    assert result == {'status': 'eventful'}


def test_dvcs_loader():
    loader = DummyDVCSLoader()
    result = loader.load()
    assert result == {'status': 'eventful'}


def test_loader_logger_default_name():
    loader = DummyBaseLoader()
    assert isinstance(loader.log, logging.Logger)
    assert loader.log.name == \
        'swh.loader.core.tests.test_loader.DummyBaseLoader'

    loader = DummyDVCSLoader()
    assert isinstance(loader.log, logging.Logger)
    assert loader.log.name == \
        'swh.loader.core.tests.test_loader.DummyDVCSLoader'


def test_loader_logger_with_name():
    loader = DummyBaseLoader('some.logger.name')
    assert isinstance(loader.log, logging.Logger)
    assert loader.log.name == \
        'some.logger.name'


@pytest.mark.fs
def test_loader_save_data_path(tmp_path):
    loader = DummyBaseLoader('some.logger.name.1')
    url = 'http://bitbucket.org/something'
    loader.origin = Origin(url=url)
    loader.visit_date = datetime.datetime(year=2019, month=10, day=1)
    loader.config = {
        'save_data_path': tmp_path,
    }

    hash_url = hashlib.sha1(url.encode('utf-8')).hexdigest()
    expected_save_path = '%s/sha1:%s/%s/2019' % (
        str(tmp_path), hash_url[0:2], hash_url
    )

    save_path = loader.get_save_data_path()
    assert save_path == expected_save_path



# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Convert objects to dictionaries suitable for swh.storage"""

import logging

from typing import Dict, Iterable, List, Optional, Tuple

from swh.model.hashutil import hash_to_hex
from swh.model.model import BaseContent, Content, SkippedContent


logger = logging.getLogger(__name__)


def prepare_contents(
        contents: Iterable[Dict], max_content_size: Optional[int] = None,
        origin_url: Optional[str] = None) -> Tuple[
            List[Dict], List[Dict]]:
    """Prepare contents for storage from a list of contents

    Returns
        tuple of content iterable, skipped content iterable

    """
    present_contents: List[Dict] = []
    skipped_contents: List[Dict] = []
    for _content in contents:
        content = content_for_storage(
            _content, max_content_size=max_content_size, origin_url=origin_url)
        if isinstance(content, SkippedContent):
            skipped_contents.append(content.to_dict())
        else:
            present_contents.append(content.to_dict())
    return present_contents, skipped_contents


def content_for_storage(
        content: Dict, max_content_size: Optional[int] = None,
        origin_url: Optional[str] = None) -> BaseContent:
    """Prepare content to be ready for storage

    Note:
    - 'data' is returned only if max_content_size is not reached.

    Returns:
        content with added data (or reason for being missing)

    """
    ret = content.copy()
    ret.pop('perms', None)

    if max_content_size and ret['length'] > max_content_size:
        logger.info('Skipping content %s, too large (%s > %s)' %
                    (hash_to_hex(content['sha1_git']),
                     ret['length'],
                     max_content_size))
        ret.pop('data', None)
        ret.update({'status': 'absent',
                    'reason': 'Content too large',
                    'origin': origin_url})
        return SkippedContent.from_dict(ret)

    if 'data' not in ret:
        with open(ret['path'], 'rb') as f:
            ret['data'] = f.read()

    # Extra keys added by swh.model.from_disk, that are not accepted
    # by swh-storage
    ret.pop('path', None)

    ret['status'] = 'visible'

    return Content.from_dict(ret)



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


def test_npm_loader(mocker, swh_app, celery_session_worker, swh_config):
    mock_loader = mocker.patch('swh.loader.package.npm.loader.NpmLoader.load')
    mock_loader.return_value = {'status': 'eventful'}

    res = swh_app.send_task(
        'swh.loader.package.npm.tasks.LoadNpm',
        kwargs=dict(url='https://www.npmjs.com/package/some-package'))
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {'status': 'eventful'}



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest


from swh.loader.cli import run, list, get_loader, SUPPORTED_LOADERS
from swh.loader.package.loader import PackageLoader

from click.testing import CliRunner


def test_get_loader_wrong_input(swh_config):
    """Unsupported loader should raise

    """
    loader_type = 'unknown'
    assert loader_type not in SUPPORTED_LOADERS
    with pytest.raises(ValueError, match='Invalid loader'):
        get_loader(loader_type, url='db-url')


def test_get_loader(swh_config):
    """Instantiating a supported loader should be ok

    """
    loader_input = {
        'archive': {
            'url': 'some-url',
            'artifacts': [],
        },
        'debian': {
            'url': 'some-url',
            'date': 'something',
            'packages': [],
        },
        'deposit': {
            'url': 'some-url',
            'deposit_id': 1,
        },
        'npm': {
            'url': 'https://www.npmjs.com/package/onepackage',
        },
        'pypi': {
            'url': 'some-url',
        },
    }
    for loader_type, kwargs in loader_input.items():
        loader = get_loader(loader_type, **kwargs)
        assert isinstance(loader, PackageLoader)


def test_run_help(swh_config):
    """Help message should be ok

    """
    runner = CliRunner()
    result = runner.invoke(run, ['-h'])

    assert result.exit_code == 0
    expected_help_msg = """Usage: run [OPTIONS] [archive|cran|debian|deposit|functional|npm|pypi] URL
           [OPTIONS]...

  Ingest with loader <type> the origin located at <url>

Options:
  -h, --help  Show this message and exit.
"""  # noqa

    assert result.output.startswith(expected_help_msg)


def test_run_pypi(mocker, swh_config):
    """Triggering a load should be ok

    """
    mock_loader = mocker.patch('swh.loader.package.pypi.loader.PyPILoader')
    runner = CliRunner()
    result = runner.invoke(run, ['pypi', 'https://some-url'])
    assert result.exit_code == 0
    mock_loader.assert_called_once_with(url='https://some-url')  # constructor


def test_list_help(mocker, swh_config):
    """Triggering a load should be ok

    """
    runner = CliRunner()
    result = runner.invoke(list, ['--help'])
    assert result.exit_code == 0
    expected_help_msg = """Usage: list [OPTIONS] [[all|archive|cran|debian|deposit|functional|npm|pypi]]

  List supported loaders and optionally their arguments

Options:
  -h, --help  Show this message and exit.
"""  # noqa
    assert result.output.startswith(expected_help_msg)


def test_list_help_npm(mocker, swh_config):
    """Triggering a load should be ok

    """
    runner = CliRunner()
    result = runner.invoke(list, ['npm'])
    assert result.exit_code == 0
    expected_help_msg = '''Loader: Load npm origin's artifact releases into swh archive.
signature: (url: str)
'''  # noqa
    assert result.output.startswith(expected_help_msg)



# Copyright (C) 2019-2020 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import pytest

from datetime import datetime, timezone
from dateutil.tz import tzlocal

from os import path

from swh.loader.package.cran.loader import (
    extract_intrinsic_metadata, CRANLoader, parse_date,
    parse_debian_control
)
from swh.core.tarball import uncompress
from swh.model.model import TimestampWithTimezone

from swh.loader.package.tests.common import (
    check_snapshot, get_stats
)


def test_cran_parse_date():
    data = [
        # parsable, some have debatable results though
        ('2001-June-08',
         datetime(2001, 6, 8, 0, 0, tzinfo=timezone.utc)),
        ('Tue Dec 27 15:06:08 PST 2011',
         datetime(2011, 12, 27, 15, 6, 8, tzinfo=timezone.utc)),
        ('8-14-2013',
         datetime(2013, 8, 14, 0, 0, tzinfo=timezone.utc)),
        ('2011-01',
         datetime(2011, 1, 1, 0, 0, tzinfo=timezone.utc)),
        ('201109',
         datetime(2009, 11, 20, 0, 0, tzinfo=timezone.utc)),
        ('04-12-2014',
         datetime(2014, 4, 12, 0, 0, tzinfo=timezone.utc)),
        ('2018-08-24, 10:40:10',
         datetime(2018, 8, 24, 10, 40, 10, tzinfo=timezone.utc)),
        ('2013-October-16',
         datetime(2013, 10, 16, 0, 0, tzinfo=timezone.utc)),
        ('Aug 23, 2013',
         datetime(2013, 8, 23, 0, 0, tzinfo=timezone.utc)),
        ('27-11-2014',
         datetime(2014, 11, 27, 0, 0, tzinfo=timezone.utc)),
        ('2019-09-26,',
         datetime(2019, 9, 26, 0, 0, tzinfo=timezone.utc)),
        ('9/25/2014',
         datetime(2014, 9, 25, 0, 0, tzinfo=timezone.utc)),
        ('Fri Jun 27 17:23:53 2014',
         datetime(2014, 6, 27, 17, 23, 53, tzinfo=timezone.utc)),
        ('28-04-2014',
         datetime(2014, 4, 28, 0, 0, tzinfo=timezone.utc)),
        ('04-14-2014',
         datetime(2014, 4, 14, 0, 0, tzinfo=timezone.utc)),
        ('2019-05-08 14:17:31 UTC',
         datetime(2019, 5, 8, 14, 17, 31, tzinfo=timezone.utc)),
        ('Wed May 21 13:50:39 CEST 2014',
         datetime(2014, 5, 21, 13, 50, 39, tzinfo=tzlocal())),
        ('2018-04-10 00:01:04 KST',
         datetime(2018, 4, 10, 0, 1, 4, tzinfo=timezone.utc)),
        ('2019-08-25 10:45',
         datetime(2019, 8, 25, 10, 45, tzinfo=timezone.utc)),
        ('March 9, 2015',
         datetime(2015, 3, 9, 0, 0, tzinfo=timezone.utc)),
        ('Aug. 18, 2012',
         datetime(2012, 8, 18, 0, 0, tzinfo=timezone.utc)),
        ('2014-Dec-17',
         datetime(2014, 12, 17, 0, 0, tzinfo=timezone.utc)),
        ('March 01, 2013',
         datetime(2013, 3, 1, 0, 0, tzinfo=timezone.utc)),
        ('2017-04-08.',
         datetime(2017, 4, 8, 0, 0, tzinfo=timezone.utc)),
        ('2014-Apr-22',
         datetime(2014, 4, 22, 0, 0, tzinfo=timezone.utc)),
        ('Mon Jan 12 19:54:04 2015',
         datetime(2015, 1, 12, 19, 54, 4, tzinfo=timezone.utc)),
        ('May 22, 2014',
         datetime(2014, 5, 22, 0, 0, tzinfo=timezone.utc)),
        ('2014-08-12 09:55:10 EDT',
         datetime(2014, 8, 12, 9, 55, 10, tzinfo=timezone.utc)),
        # unparsable
        ('Fabruary 21, 2012', None),
        ('2019-05-28"', None),
        ('2017-03-01 today', None),
        ('2016-11-0110.1093/icesjms/fsw182', None),
        ('2019-07-010', None),
        ('2015-02.23', None),
        ('20013-12-30', None),
        ('2016-08-017', None),
        ('2019-02-07l', None),
        ('2018-05-010', None),
        ('2019-09-27 KST', None),
        ('$Date$', None),
        ('2019-09-27 KST', None),
        ('2019-06-22 $Date$', None),
        ('$Date: 2013-01-18 12:49:03 -0600 (Fri, 18 Jan 2013) $', None),
        ('2015-7-013', None),
        ('2018-05-023', None),
        ("Check NEWS file for changes: news(package='simSummary')", None)
    ]
    for date, expected_date in data:
        actual_tstz = parse_date(date)
        if expected_date is None:
            assert actual_tstz is None, date
        else:
            expected_tstz = TimestampWithTimezone.from_datetime(expected_date)
            assert actual_tstz == expected_tstz, date


@pytest.mark.fs
def test_extract_intrinsic_metadata(tmp_path, datadir):
    """Parsing existing archive's PKG-INFO should yield results"""
    uncompressed_archive_path = str(tmp_path)
    # sample url
    # https://cran.r-project.org/src_contrib_1.4.0_Recommended_KernSmooth_2.22-6.tar.gz  # noqa
    archive_path = path.join(
        datadir, 'https_cran.r-project.org',
        'src_contrib_1.4.0_Recommended_KernSmooth_2.22-6.tar.gz')
    uncompress(archive_path, dest=uncompressed_archive_path)

    actual_metadata = extract_intrinsic_metadata(uncompressed_archive_path)

    expected_metadata = {
        'Package': 'KernSmooth',
        'Priority': 'recommended',
        'Version': '2.22-6',
        'Date': '2001-June-08',
        'Title': 'Functions for kernel smoothing for Wand & Jones (1995)',
        'Author': 'S original by Matt Wand.\n\tR port by  Brian Ripley <ripley@stats.ox.ac.uk>.',  # noqa
        'Maintainer': 'Brian Ripley <ripley@stats.ox.ac.uk>',
        'Description': 'functions for kernel smoothing (and density estimation)\n  corresponding to the book: \n  Wand, M.P. and Jones, M.C. (1995) "Kernel Smoothing".',  # noqa
        'License': 'Unlimited use and distribution (see LICENCE).',
        'URL': 'http://www.biostat.harvard.edu/~mwand'
    }

    assert actual_metadata == expected_metadata


@pytest.mark.fs
def test_extract_intrinsic_metadata_failures(tmp_path):
    """Parsing inexistent path/archive/PKG-INFO yield None"""
    # inexistent first level path
    assert extract_intrinsic_metadata('/something-inexistent') == {}
    # inexistent second level path (as expected by pypi archives)
    assert extract_intrinsic_metadata(tmp_path) == {}
    # inexistent PKG-INFO within second level path
    existing_path_no_pkginfo = str(tmp_path / 'something')
    os.mkdir(existing_path_no_pkginfo)
    assert extract_intrinsic_metadata(tmp_path) == {}


def test_cran_one_visit(swh_config, requests_mock_datadir):
    version = '2.22-6'
    base_url = 'https://cran.r-project.org'
    origin_url = f'{base_url}/Packages/Recommended_KernSmooth/index.html'
    artifact_url = f'{base_url}/src_contrib_1.4.0_Recommended_KernSmooth_{version}.tar.gz'  # noqa
    loader = CRANLoader(origin_url, artifacts=[{
        'url': artifact_url,
        'version': version,
    }])

    actual_load_status = loader.load()

    expected_snapshot_id = '920adcccc78aaeedd3cfa4459dd900d8c3431a21'
    assert actual_load_status == {
        'status': 'eventful',
        'snapshot_id': expected_snapshot_id
    }

    expected_snapshot = {
        'id': expected_snapshot_id,
        'branches': {
            'HEAD': {'target': f'releases/{version}', 'target_type': 'alias'},
            f'releases/{version}': {
                'target': '42bdb16facd5140424359c8ce89a28ecfa1ce603',
                'target_type': 'revision'
            }
        }
    }
    check_snapshot(expected_snapshot, loader.storage)

    origin_visit = loader.storage.origin_visit_get_latest(origin_url)
    assert origin_visit['status'] == 'full'
    assert origin_visit['type'] == 'cran'

    visit_stats = get_stats(loader.storage)
    assert {
        'content': 33,
        'directory': 7,
        'origin': 1,
        'origin_visit': 1,
        'person': 1,
        'release': 0,
        'revision': 1,
        'skipped_content': 0,
        'snapshot': 1
    } == visit_stats

    urls = [
        m.url for m in requests_mock_datadir.request_history
        if m.url.startswith(base_url)
    ]
    # visited each artifact once across 2 visits
    assert len(urls) == 1


def test_cran_2_visits_same_origin(
        swh_config, requests_mock_datadir):
    """Multiple visits on the same origin, only 1 archive fetch"""
    version = '2.22-6'
    base_url = 'https://cran.r-project.org'
    origin_url = f'{base_url}/Packages/Recommended_KernSmooth/index.html'
    artifact_url = f'{base_url}/src_contrib_1.4.0_Recommended_KernSmooth_{version}.tar.gz'  # noqa
    loader = CRANLoader(origin_url, artifacts=[{
        'url': artifact_url,
        'version': version
    }])

    # first visit
    actual_load_status = loader.load()

    expected_snapshot_id = '920adcccc78aaeedd3cfa4459dd900d8c3431a21'
    assert actual_load_status == {
        'status': 'eventful',
        'snapshot_id': expected_snapshot_id
    }

    expected_snapshot = {
        'id': expected_snapshot_id,
        'branches': {
            'HEAD': {'target': f'releases/{version}', 'target_type': 'alias'},
            f'releases/{version}': {
                'target': '42bdb16facd5140424359c8ce89a28ecfa1ce603',
                'target_type': 'revision'
            }
        }
    }
    check_snapshot(expected_snapshot, loader.storage)

    origin_visit = loader.storage.origin_visit_get_latest(origin_url)
    assert origin_visit['status'] == 'full'
    assert origin_visit['type'] == 'cran'

    visit_stats = get_stats(loader.storage)
    assert {
        'content': 33,
        'directory': 7,
        'origin': 1,
        'origin_visit': 1,
        'person': 1,
        'release': 0,
        'revision': 1,
        'skipped_content': 0,
        'snapshot': 1
    } == visit_stats

    # second visit
    actual_load_status2 = loader.load()

    assert actual_load_status2 == {
        'status': 'uneventful',
        'snapshot_id': expected_snapshot_id
    }

    origin_visit2 = loader.storage.origin_visit_get_latest(origin_url)
    assert origin_visit2['status'] == 'full'
    assert origin_visit2['type'] == 'cran'

    visit_stats2 = get_stats(loader.storage)
    visit_stats['origin_visit'] += 1
    assert visit_stats2 == visit_stats, 'same stats as 1st visit, +1 visit'

    urls = [
        m.url for m in requests_mock_datadir.request_history
        if m.url.startswith(base_url)
    ]
    assert len(urls) == 1, 'visited one time artifact url (across 2 visits)'


def test_parse_debian_control(datadir):
    description_file = os.path.join(datadir, 'description', 'acepack')

    actual_metadata = parse_debian_control(description_file)

    assert actual_metadata == {
        'Package': 'acepack',
        'Maintainer': 'Shawn Garbett',
        'Version': '1.4.1',
        'Author': 'Phil Spector, Jerome Friedman, Robert Tibshirani...',
        'Description': 'Two nonparametric methods for multiple regression...',
        'Title': 'ACE & AVAS 4 Selecting Multiple Regression Transformations',
        'License': 'MIT + file LICENSE',
        'Suggests': 'testthat',
        'Packaged': '2016-10-28 15:38:59 UTC; garbetsp',
        'Repository': 'CRAN',
        'Date/Publication': '2016-10-29 00:11:52',
        'NeedsCompilation': 'yes'
    }


def test_parse_debian_control_unicode_issue(datadir):
    # iso-8859-1 caused failure, now fixed
    description_file = os.path.join(
        datadir, 'description', 'KnownBR'
    )

    actual_metadata = parse_debian_control(description_file)

    assert actual_metadata == {
        'Package': 'KnowBR',
        'Version': '2.0',
        'Title': '''Discriminating Well Surveyed Spatial Units from Exhaustive
        Biodiversity Databases''',
        'Author': 'Cástor Guisande González and Jorge M. Lobo',
        'Maintainer': 'Cástor Guisande González <castor@email.es>',
        'Description':
        'It uses species accumulation curves and diverse estimators...',
        'License': 'GPL (>= 2)',
        'Encoding': 'latin1',
        'Depends': 'R (>= 3.0), fossil, mgcv, plotrix, sp, vegan',
        'Suggests': 'raster, rgbif',
        'NeedsCompilation': 'no',
        'Packaged': '2019-01-30 13:27:29 UTC; castor',
        'Repository': 'CRAN',
        'Date/Publication': '2019-01-31 20:53:50 UTC'
    }
